@@ -6,6 +6,10 @@
     it under the terms of the GNU General Public License version 2 as published 
     by the Free Software Foundation
 
+    Additionally, the license of this program contains a special
+    exception which allows to distribute the program in binary form when
+    it is linked against OpenSSL.
+
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@@ -17,7 +21,16 @@
 */
 #include "arguments.hh"
 #include <boost/algorithm/string.hpp>
+#include <boost/algorithm/string/compare.hpp>
+#include <boost/algorithm/string/predicate.hpp>
+#include <boost/foreach.hpp>
 #include "namespaces.hh"
+#include "logger.hh"
+#include <sys/types.h>
+#include <dirent.h>
+#include <sys/stat.h>
+#include <unistd.h>
+#include <limits.h>
 
 const ArgvMap::param_t::const_iterator ArgvMap::begin()
 {
@@ -129,41 +142,42 @@ string ArgvMap::helpstring(string prefix)
   return help;
 }
 
-string ArgvMap::configstring()
+string ArgvMap::configstring(bool current)
 {
   string help;
+
+  if (current)
+    help="# Autogenerated configuration file based on running instance\n";
+  else
+    help="# Autogenerated configuration file template\n";
   
-  help="# Autogenerated configuration file template\n";
-  for(map<string,string>::const_iterator i=helpmap.begin();
-      i!=helpmap.end();
-      i++)
-    {
-      if(d_typeMap[i->first]=="Command")
-        continue;
+  for(map<string,string>::const_iterator i=helpmap.begin(); i!=helpmap.end(); i++) {
+    if(d_typeMap[i->first]=="Command")
+      continue;
 
-      help+="#################################\n";
-      help+="# ";
-      help+=i->first;
-      help+="\t";
-      help+=i->second;
-      help+="\n#\n";
+    help+="#################################\n";
+    help+="# ";
+    help+=i->first;
+    help+="\t";
+    help+=i->second;
+    help+="\n#\n";
+    if (current) {
+      help+=i->first+"="+params[i->first]+"\n\n";
+    } else {
       help+="# "+i->first+"="+params[i->first]+"\n\n";
-
     }
+  }
   return help;
 }
 
-
 const string & ArgvMap::operator[](const string &arg)
 {
   if(!parmIsset(arg))
     throw ArgException(string("Undefined but needed argument: '")+arg+"'");
 
-
   return params[arg];
 }
 
-#ifndef WIN32
 mode_t ArgvMap::asMode(const string &arg) 
 {
   mode_t mode;
@@ -221,7 +235,6 @@ uid_t ArgvMap::asUid(const string &arg)
    }
    return uid;
 }
-#endif
 
 int ArgvMap::asNum(const string &arg)
 {
@@ -286,37 +299,59 @@ void ArgvMap::parseOne(const string &arg, const string &parseOnly, bool lax)
 {
   string var, val;
   string::size_type pos;
+  bool incremental = false;
 
-  if(!arg.find("--") &&(pos=arg.find("="))!=string::npos)  // this is a --port=25 case
-    {
-      var=arg.substr(2,pos-2);
-      val=arg.substr(pos+1);
-    }
-  else if(!arg.find("--") && (arg.find("=")==string::npos))  // this is a --daemon case
-    { 
-      var=arg.substr(2);
-      val="";
-    }
-  else if(arg[0]=='-')
-    {
-      var=arg.substr(1);
-      val="";
-    }
-  else { // command 
-    d_cmds.push_back(arg);
+  if(!arg.find("--") && (pos=arg.find("+="))!=string::npos) // this is a --port+=25 case
+  {
+    var=arg.substr(2,pos-2);
+    val=arg.substr(pos+2);
+    incremental = true;
   }
+  else if(!arg.find("--") && (pos=arg.find("="))!=string::npos)  // this is a --port=25 case
+  {
+    var=arg.substr(2,pos-2);
+    val=arg.substr(pos+1);
+  }
+  else if(!arg.find("--") && (arg.find("=")==string::npos))  // this is a --daemon case
+  {
+    var=arg.substr(2);
+    val="";
+  }
+  else if(arg[0]=='-')
+  {
+    var=arg.substr(1);
+    val="";
+  }
+  else // command
+    d_cmds.push_back(arg);
+ 
+  boost::trim(var);
 
   if(var!="" && (parseOnly.empty() || var==parseOnly)) {
-
     pos=val.find_first_not_of(" \t");  // strip leading whitespace
-    if(pos && pos!=string::npos) 
+    if(pos && pos!=string::npos)
       val=val.substr(pos);
-
     if(parmIsset(var))
-      params[var]=val;
-    else
-      if(!lax)
-        throw ArgException("Trying to set unexisting parameter '"+var+"'");
+    {
+      if(incremental)
+      {
+        if(params[var].empty())
+        {
+          if(!d_cleared.count(var))
+            throw ArgException("Incremental parameter '"+var+"' without a parent");
+          params[var]=val;
+        }
+        else
+          params[var]+=", " + val;
+      }
+      else
+      {
+        params[var]=val;
+        d_cleared.insert(var);
+      }
+    }
+    else if(!lax)
+      throw ArgException("Trying to set unknown parameter '"+var+"'");
   }
 }
 
@@ -328,6 +363,7 @@ const vector<string>&ArgvMap::getCommands()
 void ArgvMap::parse(int &argc, char **argv, bool lax)
 {
   d_cmds.clear();
+  d_cleared.clear();
   for(int n=1;n<argc;n++) {
     parseOne(argv[n],"",lax);
   }
@@ -384,14 +420,21 @@ bool ArgvMap::preParseFile(const char *fname, const string &arg, const string& t
   return true;
 }
 
-
 bool ArgvMap::file(const char *fname, bool lax)
+{
+   return file(fname,lax,false);
+}
+
+bool ArgvMap::file(const char *fname, bool lax, bool included)
 {
   ifstream f(fname);
   if(!f) {
     return false;
   }
 
+  if (!parmIsset("include-dir"))  // inject include-dir
+    set("include-dir","Directory to include configuration files from");
+
   string line;
   string pline;
   string::size_type pos;
@@ -420,5 +463,59 @@ bool ArgvMap::file(const char *fname, bool lax)
     line="";
   }
 
+  // handle include here (avoid re-include)
+  if (!included && !params["include-dir"].empty()) {
+    std::vector<std::string> extraConfigs;
+    gatherIncludes(extraConfigs); 
+    BOOST_FOREACH(const std::string& fn, extraConfigs) {
+      if (!file(fn.c_str(), lax, true)) {
+        L << Logger::Error << fn << " could not be parsed" << std::endl;
+        throw ArgException(fn + " could not be parsed");
+      }
+    }
+  }
+
   return true;
 }
+
+void ArgvMap::gatherIncludes(std::vector<std::string> &extraConfigs) {
+  extraConfigs.clear();
+  if (params["include-dir"].empty()) return; // nothing to do
+    struct stat st;
+    DIR *dir;
+    struct dirent *ent;
+
+    // stat
+    if (stat(params["include-dir"].c_str(), &st)) {
+       L << Logger::Error << params["include-dir"] << " does not exist!" << std::endl;
+       throw ArgException(params["include-dir"] + " does not exist!");
+    }
+
+    // wonder if it's accessible directory
+    if (!S_ISDIR(st.st_mode)) {
+       L << Logger::Error << params["include-dir"] << " is not a directory" << std::endl;
+       throw ArgException(params["include-dir"] + " is not a directory");
+    }
+
+    if (!(dir = opendir(params["include-dir"].c_str()))) {
+       L << Logger::Error << params["include-dir"] << " is not accessible" << std::endl;
+       throw ArgException(params["include-dir"] + " is not accessible");
+    }
+
+    while((ent = readdir(dir)) != NULL) {
+      if (ent->d_name[0] == '.') continue; // skip any dots
+      if (boost::ends_with(ent->d_name, ".conf")) {
+        // build name
+        std::ostringstream namebuf;
+        namebuf << params["include-dir"].c_str() << "/" << ent->d_name; // FIXME: Use some path separator
+        // ensure it's readable file
+        if (stat(namebuf.str().c_str(), &st) || !S_ISREG(st.st_mode)) {
+          L << Logger::Error << namebuf.str() << " is not a file" << std::endl;
+          throw ArgException(namebuf.str() + " does not exist!");
+        }
+        extraConfigs.push_back(namebuf.str());
+      }
+    }
+    std::sort(extraConfigs.begin(), extraConfigs.end(), CIStringComparePOSIX()); 
+    closedir(dir);
+}
