@@ -46,6 +46,7 @@ DNSPacketWriter::DNSPacketWriter(vector<uint8_t>& content, const string& qname,
 
   d_stuff=0xffff;
   d_labelmap.reserve(16);
+  d_truncatemarker=d_content.size();
 }
 
 dnsheader* DNSPacketWriter::getHeader()
@@ -53,7 +54,7 @@ dnsheader* DNSPacketWriter::getHeader()
   return (dnsheader*)&*d_content.begin();
 }
 
-void DNSPacketWriter::startRecord(const string& name, uint16_t qtype, uint32_t ttl, uint16_t qclass, Place place)
+void DNSPacketWriter::startRecord(const string& name, uint16_t qtype, uint32_t ttl, uint16_t qclass, Place place, bool compress)
 {
   if(!d_record.empty()) 
     commit();
@@ -64,19 +65,19 @@ void DNSPacketWriter::startRecord(const string& name, uint16_t qtype, uint32_t t
   d_recordttl=ttl;
   d_recordplace=place;
 
-  d_stuff = 0; 
+  d_stuff = 0;
   d_rollbackmarker=d_content.size();
 
-  if(pdns_iequals(d_qname, d_recordqname)) {  // don't do the whole label compression thing if we *know* we can get away with "see question"
-    static char marker[2]={0xc0, 0x0c};
-    d_content.insert(d_content.end(), &marker[0], &marker[2]);
+  if(compress && pdns_iequals(d_qname, d_recordqname)) {  // don't do the whole label compression thing if we *know* we can get away with "see question"
+    static unsigned char marker[2]={0xc0, 0x0c};
+    d_content.insert(d_content.end(), (const char *) &marker[0], (const char *) &marker[2]);
   }
   else {
-    xfrLabel(d_recordqname, true);
+    xfrLabel(d_recordqname, compress);
     d_content.insert(d_content.end(), d_record.begin(), d_record.end());
     d_record.clear();
   }
-      
+
   d_stuff = sizeof(dnsrecordheader); // this is needed to get compressed label offsets right, the dnsrecordheader will be interspersed
   d_sor=d_content.size() + d_stuff; // start of real record 
 }
@@ -237,6 +238,7 @@ void DNSPacketWriter::xfrLabel(const string& Label, bool compress)
 
     if(unescaped) {
       string part(label.c_str() + i -> first, i->second - i->first);
+      // FIXME: this relies on the semi-canonical escaped output from getLabelFromContent
       boost::replace_all(part, "\\.", ".");
       boost::replace_all(part, "\\032", " ");
       boost::replace_all(part, "\\\\", "\\"); 
@@ -247,7 +249,7 @@ void DNSPacketWriter::xfrLabel(const string& Label, bool compress)
       d_record.resize(len + part.size());
 
       memcpy(((&*d_record.begin()) + len), part.c_str(), part.size());
-      pos+=(part.size())+1;        		 
+      pos+=(part.size())+1;                         
     }
     else {
       char labelsize=(char)(i->second - i->first);
@@ -293,6 +295,15 @@ void DNSPacketWriter::rollback()
   d_stuff=0;
 }
 
+void DNSPacketWriter::truncate()
+{
+  d_content.resize(d_truncatemarker);
+  d_record.clear();
+  d_stuff=0;
+  dnsheader* dh=reinterpret_cast<dnsheader*>( &*d_content.begin());
+  dh->ancount = dh->nscount = dh->arcount = 0;
+}
+
 void DNSPacketWriter::commit()
 {
   if(d_stuff==0xffff && (d_content.size()!=d_sor || !d_record.empty()))
