@@ -1,11 +1,14 @@
 /*
     PowerDNS Versatile Database Driven Nameserver
-    Copyright (C) 2002 - 2010  PowerDNS.COM BV
+    Copyright (C) 2002 - 2014  PowerDNS.COM BV
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License version 2
     as published by the Free Software Foundation
-    
+
+    Additionally, the license of this program contains a special
+    exception which allows to distribute the program in binary form when
+    it is linked against OpenSSL.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -17,15 +20,14 @@
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
-#ifndef WIN32
 #include <sys/param.h>
 #include <netdb.h>
 #include <sys/time.h>
 #include <time.h>
+#include <sys/resource.h>
 #include <netinet/in.h>
+#include <sys/un.h>
 #include <unistd.h>
-#endif // WIN32
-
 #include "misc.hh"
 #include <vector>
 #include <sstream>
@@ -39,10 +41,12 @@
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
-#include "ahuexception.hh"
+#include "pdnsexception.hh"
 #include <sys/types.h>
 #include "utility.hh"
 #include <boost/algorithm/string.hpp>
+#include "logger.hh"
+#include "iputils.hh"
 
 bool g_singleThreaded;
 
@@ -69,6 +73,28 @@ int writen2(int fd, const void *buf, size_t count)
   return count;
 }
 
+int readn2(int fd, void* buffer, unsigned int len)
+{
+  unsigned int pos=0;
+  int res;
+  for(;;) {
+    res = read(fd, (char*)buffer + pos, len - pos);
+    if(res == 0) 
+      throw runtime_error("EOF while writing message");
+    if(res < 0) {
+      if (errno == EAGAIN)
+        throw std::runtime_error("used writen2 on non-blocking socket, got EAGAIN");
+      else
+        unixDie("failed in writen2");
+    } 
+    
+    pos+=res;
+    if(pos == len)
+      break;
+  }
+  return len;
+}
+
 
 string nowTime()
 {
@@ -213,53 +239,12 @@ bool dottedEndsOn(const string &domain, const string &suffix)
   return true;
 }
 
-int sendData(const char *buffer, int replen, int outsock)
-{
-  uint16_t nlen=htons(replen);
-  Utility::iovec iov[2];
-  iov[0].iov_base=(char*)&nlen;
-  iov[0].iov_len=2;
-  iov[1].iov_base=(char*)buffer;
-  iov[1].iov_len=replen;
-  int ret=Utility::writev(outsock,iov,2);
-
-  if(ret <= 0)  // "EOF is error" - we can't deal with EAGAIN errors at this stage yet
-    return -1;
-
-  if(ret!=replen+2) {
-    // we can safely assume ret > 2, as 2 is < PIPE_BUF
-    
-    buffer += (ret - 2);
-    replen -= (ret - 2);
-
-    while (replen) {
-      ret = write(outsock, buffer, replen);
-      if(ret < 0) {
-        if(errno==EAGAIN) { // wait, we might've exhausted the window
-          while(waitForRWData(outsock, false, 1, 0)==0)
-            ;
-          continue;
-        }
-        return ret;
-      }
-      if(!ret)
-        return -1; // "EOF == error"
-      replen -= ret;
-      buffer += ret;
-    }
-    if(!replen)
-      return 0;
-    return -1;
-  }
-  return 0;
-}
-
 static void parseService4(const string &descr, ServiceTuple &st)
 {
   vector<string>parts;
   stringtok(parts,descr,":");
   if(parts.empty())
-    throw AhuException("Unable to parse '"+descr+"' as a service");
+    throw PDNSException("Unable to parse '"+descr+"' as a service");
   st.host=parts[0];
   if(parts.size()>1)
     st.port=atoi(parts[1].c_str());
@@ -269,7 +254,7 @@ static void parseService6(const string &descr, ServiceTuple &st)
 {
   string::size_type pos=descr.find(']');
   if(pos == string::npos)
-    throw AhuException("Unable to parse '"+descr+"' as an IPv6 service");
+    throw PDNSException("Unable to parse '"+descr+"' as an IPv6 service");
 
   st.host=descr.substr(1, pos-1);
   if(pos + 2 < descr.length())
@@ -280,7 +265,7 @@ static void parseService6(const string &descr, ServiceTuple &st)
 void parseService(const string &descr, ServiceTuple &st)
 {
   if(descr.empty())
-    throw AhuException("Unable to parse '"+descr+"' as a service");
+    throw PDNSException("Unable to parse '"+descr+"' as a service");
 
   vector<string> parts;
   stringtok(parts, descr, ":");
@@ -424,11 +409,8 @@ string urlEncode(const string &text)
 
 string getHostname()
 {
-#ifdef WIN32
-# define MAXHOSTNAMELEN 1025
-#endif // WIN32
 #ifndef MAXHOSTNAMELEN
-# define MAXHOSTNAMELEN 255
+#define MAXHOSTNAMELEN 255
 #endif
 
   char tmp[MAXHOSTNAMELEN];
@@ -458,25 +440,10 @@ string stringerror()
   return strerror(errno);
 }
 
-#ifdef WIN32
-string netstringerror()
-{
-  char buf[512];
-  int err=WSAGetLastError();
-  if(FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM, NULL, err,
-        	     0, buf, sizeof(buf)-1, NULL)) {
-    return string(buf);
-  }
-  else {
-    return strerror(err);
-  }
-}
-#else
 string netstringerror()
 {
   return stringerror();
 }
-#endif
 
 void cleanSlashes(string &str)
 {
@@ -705,33 +672,30 @@ int makeIPv6sockaddr(const std::string& addr, struct sockaddr_in6* ret)
     port = atoi(addr.c_str()+pos+2);  
   }
   
-  struct addrinfo* res;
-  struct addrinfo hints;
-  memset(&hints, 0, sizeof(hints));
+  if(inet_pton(AF_INET6, ourAddr.c_str(), (void*)&ret->sin6_addr) != 1) {
+    struct addrinfo* res;
+    struct addrinfo hints;
+    memset(&hints, 0, sizeof(hints));
+    
+    hints.ai_family = AF_INET6;
+    hints.ai_flags = AI_NUMERICHOST;
+    
+    int error;
+    if((error=getaddrinfo(ourAddr.c_str(), 0, &hints, &res))) { // this is correct
+      return -1;
+    }
   
-  hints.ai_family = AF_INET6;
-  hints.ai_flags = AI_NUMERICHOST;
-  
-  int error;
-  if((error=getaddrinfo(ourAddr.c_str(), 0, &hints, &res))) { // this is correct
-    /*
-    cerr<<"Error translating IPv6 address '"<<addr<<"': ";
-    if(error==EAI_SYSTEM)
-      cerr<<strerror(errno)<<endl;
-    else
-      cerr<<gai_strerror(error)<<endl;
-    */
-    return -1;
+    memcpy(ret, res->ai_addr, res->ai_addrlen);
+    freeaddrinfo(res);
   }
-  
-  memcpy(ret, res->ai_addr, res->ai_addrlen);
+
   if(port >= 0)
     ret->sin6_port = htons(port);
-  freeaddrinfo(res);
+
   return 0;
 }
 
-int makeIPv4sockaddr(const string &str, struct sockaddr_in* ret)
+int makeIPv4sockaddr(const std::string& str, struct sockaddr_in* ret)
 {
   if(str.empty()) {
     return -1;
@@ -762,6 +726,19 @@ int makeIPv4sockaddr(const string &str, struct sockaddr_in* ret)
   return -1;
 }
 
+int makeUNsockaddr(const std::string& path, struct sockaddr_un* ret)
+{
+  if (path.empty())
+    return -1;
+
+  memset(ret, 0, sizeof(struct sockaddr_un));
+  ret->sun_family = AF_UNIX;
+  if (path.length() >= sizeof(ret->sun_path))
+    return -1;
+
+  path.copy(ret->sun_path, sizeof(ret->sun_path), 0);
+  return 0;
+}
 
 //! read a line of text from a FILE* to a std::string, returns false on 'no data'
 bool stringfgets(FILE* fp, std::string& line)
@@ -778,8 +755,146 @@ bool stringfgets(FILE* fp, std::string& line)
   return true;
 }
 
+bool readFileIfThere(const char* fname, std::string* line)
+{
+  line->clear();
+  FILE* fp = fopen(fname, "r");
+  if(!fp)
+    return false;
+  stringfgets(fp, *line);
+  fclose(fp);
+  return true;
+}
+
 Regex::Regex(const string &expr)
 {
   if(regcomp(&d_preg, expr.c_str(), REG_ICASE|REG_NOSUB|REG_EXTENDED))
-    throw AhuException("Regular expression did not compile");
+    throw PDNSException("Regular expression did not compile");
+}
+
+void addCMsgSrcAddr(struct msghdr* msgh, void* cmsgbuf, ComboAddress* source)
+{
+  struct cmsghdr *cmsg = NULL;
+
+  if(source->sin4.sin_family == AF_INET6) {
+    struct in6_pktinfo *pkt;
+
+    msgh->msg_control = cmsgbuf;
+    msgh->msg_controllen = CMSG_SPACE(sizeof(*pkt));
+
+    cmsg = CMSG_FIRSTHDR(msgh);
+    cmsg->cmsg_level = IPPROTO_IPV6;
+    cmsg->cmsg_type = IPV6_PKTINFO;
+    cmsg->cmsg_len = CMSG_LEN(sizeof(*pkt));
+
+    pkt = (struct in6_pktinfo *) CMSG_DATA(cmsg);
+    memset(pkt, 0, sizeof(*pkt));
+    pkt->ipi6_addr = source->sin6.sin6_addr;
+    msgh->msg_controllen = cmsg->cmsg_len; // makes valgrind happy and is slightly better style
+  }
+  else {
+#ifdef IP_PKTINFO
+    struct in_pktinfo *pkt;
+
+    msgh->msg_control = cmsgbuf;
+    msgh->msg_controllen = CMSG_SPACE(sizeof(*pkt));
+
+    cmsg = CMSG_FIRSTHDR(msgh);
+    cmsg->cmsg_level = IPPROTO_IP;
+    cmsg->cmsg_type = IP_PKTINFO;
+    cmsg->cmsg_len = CMSG_LEN(sizeof(*pkt));
+
+    pkt = (struct in_pktinfo *) CMSG_DATA(cmsg);
+    memset(pkt, 0, sizeof(*pkt));
+    pkt->ipi_spec_dst = source->sin4.sin_addr;
+    msgh->msg_controllen = cmsg->cmsg_len;
+#endif
+#ifdef IP_SENDSRCADDR
+    struct in_addr *in;
+
+    msgh->msg_control = cmsgbuf;
+    msgh->msg_controllen = CMSG_SPACE(sizeof(*in));
+
+    cmsg = CMSG_FIRSTHDR(msgh);
+    cmsg->cmsg_level = IPPROTO_IP;
+    cmsg->cmsg_type = IP_SENDSRCADDR;
+    cmsg->cmsg_len = CMSG_LEN(sizeof(*in));
+
+    in = (struct in_addr *) CMSG_DATA(cmsg);
+    *in = source->sin4.sin_addr;
+    msgh->msg_controllen = cmsg->cmsg_len;
+#endif
+  }
+}
+
+unsigned int getFilenumLimit(bool hardOrSoft)
+{
+  struct rlimit rlim;
+  if(getrlimit(RLIMIT_NOFILE, &rlim) < 0)
+    unixDie("Requesting number of available file descriptors");
+  return hardOrSoft ? rlim.rlim_max : rlim.rlim_cur;
+}
+
+void setFilenumLimit(unsigned int lim)
+{
+  struct rlimit rlim;
+
+  if(getrlimit(RLIMIT_NOFILE, &rlim) < 0)
+    unixDie("Requesting number of available file descriptors");
+  rlim.rlim_cur=lim;
+  if(setrlimit(RLIMIT_NOFILE, &rlim) < 0)
+    unixDie("Setting number of available file descriptors");
+}
+
+#define burtlemix(a,b,c) \
+{ \
+  a -= b; a -= c; a ^= (c>>13); \
+  b -= c; b -= a; b ^= (a<<8); \
+  c -= a; c -= b; c ^= (b>>13); \
+  a -= b; a -= c; a ^= (c>>12);  \
+  b -= c; b -= a; b ^= (a<<16); \
+  c -= a; c -= b; c ^= (b>>5); \
+  a -= b; a -= c; a ^= (c>>3);  \
+  b -= c; b -= a; b ^= (a<<10); \
+  c -= a; c -= b; c ^= (b>>15); \
+}
+
+uint32_t burtle(const unsigned char* k, uint32_t length, uint32_t initval)
+{
+  uint32_t a,b,c,len;
+
+   /* Set up the internal state */
+  len = length;
+  a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
+  c = initval;         /* the previous hash value */
+
+  /*---------------------------------------- handle most of the key */
+  while (len >= 12) {
+    a += (k[0] +((uint32_t)k[1]<<8) +((uint32_t)k[2]<<16) +((uint32_t)k[3]<<24));
+    b += (k[4] +((uint32_t)k[5]<<8) +((uint32_t)k[6]<<16) +((uint32_t)k[7]<<24));
+    c += (k[8] +((uint32_t)k[9]<<8) +((uint32_t)k[10]<<16)+((uint32_t)k[11]<<24));
+    burtlemix(a,b,c);
+    k += 12; len -= 12;
+  }
+
+  /*------------------------------------- handle the last 11 bytes */
+  c += length;
+  switch(len) {             /* all the case statements fall through */
+  case 11: c+=((uint32_t)k[10]<<24);
+  case 10: c+=((uint32_t)k[9]<<16);
+  case 9 : c+=((uint32_t)k[8]<<8);
+    /* the first byte of c is reserved for the length */
+  case 8 : b+=((uint32_t)k[7]<<24);
+  case 7 : b+=((uint32_t)k[6]<<16);
+  case 6 : b+=((uint32_t)k[5]<<8);
+  case 5 : b+=k[4];
+  case 4 : a+=((uint32_t)k[3]<<24);
+  case 3 : a+=((uint32_t)k[2]<<16);
+  case 2 : a+=((uint32_t)k[1]<<8);
+  case 1 : a+=k[0];
+    /* case 0: nothing left to add */
+  }
+  burtlemix(a,b,c);
+  /*-------------------------------------------- report the result */
+  return c;
 }
