@@ -1,11 +1,15 @@
 /*
     PowerDNS Versatile Database Driven Nameserver
-    Copyright (C) 2005 - 2007 PowerDNS.COM BV
+    Copyright (C) 2005 - 2011 PowerDNS.COM BV
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License version 2 as 
     published by the Free Software Foundation
 
+    Additionally, the license of this program contains a special
+    exception which allows to distribute the program in binary form when
+    it is linked against OpenSSL.
+
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@@ -22,6 +26,7 @@
 #include <boost/lexical_cast.hpp>
 #include <boost/algorithm/string.hpp>
 #include <iostream>
+#include "base32.hh"
 #include "base64.hh"
 #include "namespaces.hh"
 
@@ -77,7 +82,7 @@ void RecordTextReader::xfrTime(uint32_t &val)
 
   tm.tm_year-=1900;
   tm.tm_mon-=1;
-//  val=(uint32_t)timegm(&tm); XXX FIXME disabled for Solaris
+  val=(uint32_t)Utility::timegm(&tm); 
 }
 
 void RecordTextReader::xfrIP(uint32_t &val)
@@ -123,6 +128,33 @@ void RecordTextReader::xfrIP(uint32_t &val)
 }
 
 
+void RecordTextReader::xfrIP6(std::string &val)
+{
+  struct in6_addr tmpbuf;
+
+  skipSpaces();
+  
+  size_t len;
+  // lookup end of value - think of ::ffff encoding too, has dots in it!
+  for(len=0; 
+      d_pos+len < d_string.length() && (isxdigit(d_string.at(d_pos+len)) || d_string.at(d_pos+len) == ':' || d_string.at(d_pos+len)=='.');
+    len++);
+
+  if(!len)
+    throw RecordTextException("while parsing IPv6 address, expected xdigits at position "+lexical_cast<string>(d_pos)+" in '"+d_string+"'");
+
+  // end of value is here, try parse as IPv6
+  string address=d_string.substr(d_pos, len);
+  
+  if (inet_pton(AF_INET6, address.c_str(), &tmpbuf) != 1) {
+    throw RecordTextException("while parsing IPv6 address: '" + address + "' is invalid");
+  }
+
+  val = std::string((char*)tmpbuf.s6_addr, 16);
+
+  d_pos += len;
+}
+
 bool RecordTextReader::eof()
 {
   return d_pos==d_end;
@@ -146,6 +178,7 @@ void RecordTextReader::xfr8BitInt(uint8_t &val)
     throw RecordTextException("Overflow reading 8 bit integer from record content"); // fixme improve
 }
 
+// this code should leave all the escapes around 
 void RecordTextReader::xfrLabel(string& val, bool) 
 {
   skipSpaces();
@@ -157,13 +190,9 @@ void RecordTextReader::xfrLabel(string& val, bool)
   while(d_pos < d_end) {
     if(strptr[d_pos]!='\r' && dns_isspace(strptr[d_pos]))
       break;
-
-    if(strptr[d_pos]=='\\' && d_pos < d_end - 1 && strptr[d_pos+1]!='.')  // leave the \. escape around
-      d_pos++;
-
+      
     d_pos++;
   }
-
   val.append(strptr+begin_pos, strptr+d_pos);      
 
   if(val.empty())
@@ -222,28 +251,63 @@ static inline uint8_t hextodec(uint8_t val)
 }
 
 
-void HEXDecode(const char* begin, const char* end, string& val)
+void HEXDecode(const char* begin, const char* end, string& out)
 {
-  if((end - begin)%2)
-    throw RecordTextException("Hexadecimal blob with odd number of characters");
-
-  int limit=(int)(end-begin)/2;
-  val.resize(limit);
-  for(int n=0; n < limit; ++n) {
-    val[n] = hextodec(begin[2*n])*16 + hextodec(begin[2*n+1]); 
+  if(end - begin == 1 && *begin=='-') {
+    out.clear();
+    return;
   }
+  out.clear();
+  out.reserve((end-begin)/2);
+  uint8_t mode=0, val=0;
+  for(; begin != end; ++begin) {
+    if(!isalnum(*begin))
+      continue;
+    if(mode==0) {
+      val = 16*hextodec(*begin);
+      mode=1;
+    } else {
+      val += hextodec(*begin); 
+      out.append(1, (char) val);
+      mode = 0;
+      val = 0;
+    }
+  }
+  if(mode)
+    out.append(1, (char) val);
+
 }
 
-void RecordTextReader::xfrHexBlob(string& val)
+void RecordTextReader::xfrHexBlob(string& val, bool keepReading)
+{
+  skipSpaces();
+  int pos=(int)d_pos;
+  while(d_pos < d_end && (keepReading || !dns_isspace(d_string[d_pos])))
+    d_pos++;
+
+  HEXDecode(d_string.c_str()+pos, d_string.c_str() + d_pos, val);
+}
+
+void RecordTextReader::xfrBase32HexBlob(string& val)
 {
   skipSpaces();
   int pos=(int)d_pos;
   while(d_pos < d_end && !dns_isspace(d_string[d_pos]))
     d_pos++;
 
-  HEXDecode(d_string.c_str()+pos, d_string.c_str() + d_pos, val);
+  val=fromBase32Hex(string(d_string.c_str()+pos, d_pos-pos));
 }
 
+
+void RecordTextWriter::xfrBase32HexBlob(const string& val)
+{
+  if(!d_string.empty())
+    d_string.append(1,' ');
+
+  d_string.append(toUpper(toBase32Hex(val)));
+}
+
+
 void RecordTextReader::xfrText(string& val, bool multi)
 {
   val.clear();
@@ -254,9 +318,19 @@ void RecordTextReader::xfrText(string& val, bool multi)
       val.append(1, ' ');
 
     skipSpaces();
-    if(d_string[d_pos]!='"')
+    if(d_string[d_pos]!='"') { // special case 'plenus' - without quotes
+      string::size_type pos = d_pos;
+      while(pos != d_end && isalnum(d_string[pos]))
+        pos++;
+      if(pos == d_end) {
+        val.append(1, '"');
+        val.append(d_string.c_str() + d_pos, d_end - d_pos);
+        val.append(1, '"');
+        d_pos = d_end;
+        break;
+      }
       throw RecordTextException("Data field in DNS should start with quote (\") at position "+lexical_cast<string>(d_pos)+" of '"+d_string+"'");
-
+    }
     val.append(1, '"');
     while(++d_pos < d_end && d_string[d_pos]!='"') {
       if(d_string[d_pos]=='\\' && d_pos+1!=d_end) {
@@ -357,6 +431,21 @@ void RecordTextWriter::xfrIP(const uint32_t& val)
   d_string.append(tmp, pos);
 }
 
+void RecordTextWriter::xfrIP6(const std::string& val)
+{
+  char tmpbuf[16];
+  char addrbuf[40];
+
+  if(!d_string.empty())
+   d_string.append(1,' ');
+  
+  val.copy(tmpbuf,16);
+
+  if (inet_ntop(AF_INET6, tmpbuf, addrbuf, sizeof addrbuf) == NULL)
+    throw RecordTextException("Unable to convert to ipv6 address");
+  
+  d_string += std::string(addrbuf);
+}
 
 void RecordTextWriter::xfrTime(const uint32_t& val)
 {
@@ -365,16 +454,8 @@ void RecordTextWriter::xfrTime(const uint32_t& val)
   
   struct tm tm;
   time_t time=val; // Y2038 bug!
-#ifndef WIN32
-  gmtime_r(&time, &tm);
-#else
-  struct tm* tmptr;
-  tmptr=gmtime(&time);
-  if(!tmptr)
-    throw RecordTextException("Unable to convert timestamp into pretty printable time");
-  tm=*tmptr;
-#endif
-  
+  Utility::gmtime_r(&time, &tm);
+
   char tmp[16];
   snprintf(tmp,sizeof(tmp)-1, "%04d%02d%02d" "%02d%02d%02d", 
            tm.tm_year+1900, tm.tm_mon+1, tm.tm_mday, 
@@ -394,24 +475,13 @@ void RecordTextWriter::xfr8BitInt(const uint8_t& val)
   xfr32BitInt(val);
 }
 
-
+// should not mess with the escapes
 void RecordTextWriter::xfrLabel(const string& val, bool)
 {
   if(!d_string.empty())
     d_string.append(1,' ');
-  if(val.find(' ')==string::npos) 
-    d_string+=val;
-  else {
-    d_string.reserve(d_string.size()+val.size()+3);
-    for(string::size_type pos=0; pos < val.size() ; ++pos)
-      if(dns_isspace(val[pos]))
-        d_string+="\\ ";
-      else if(val[pos]=='\\')
-        d_string.append(1,'\\');
-      else
-        d_string.append(1,val[pos]);
-  }
-  //  d_string.append(1,'.');
+  
+  d_string+=val;
 }
 
 void RecordTextWriter::xfrBlob(const string& val, int)
@@ -422,11 +492,16 @@ void RecordTextWriter::xfrBlob(const string& val, int)
   d_string+=Base64Encode(val);
 }
 
-void RecordTextWriter::xfrHexBlob(const string& val)
+void RecordTextWriter::xfrHexBlob(const string& val, bool)
 {
   if(!d_string.empty())
     d_string.append(1,' ');
 
+  if(val.empty()) {
+    d_string.append(1,'-');
+    return;
+  }
+
   string::size_type limit=val.size();
   char tmp[5];
   for(string::size_type n = 0; n < limit; ++n) {
