@@ -18,13 +18,14 @@
 #include "logger.hh"
 #include "dnsparser.hh"
 #include "arguments.hh"
-#ifndef WIN32
 #include <sys/resource.h>
 #include <sys/time.h>
-#endif
+#include "lock.hh"
+#include "responsestats.hh"
 
-using namespace std;
 #include "namespaces.hh"
+pthread_mutex_t g_carbon_config_lock=PTHREAD_MUTEX_INITIALIZER;
+
 map<string, const uint32_t*> d_get32bitpointers;
 map<string, const uint64_t*> d_get64bitpointers;
 map<string, function< uint32_t() > >  d_get32bitmembers;
@@ -56,20 +57,35 @@ optional<uint64_t> get(const string& name)
   return ret;
 }
 
-string getAllStats()
+map<string,string> getAllStatsMap()
 {
-  string ret;
+  map<string,string> ret;
+  
   pair<string, const uint32_t*> the32bits;
   pair<string, const uint64_t*> the64bits;
   pair<string, function< uint32_t() > >  the32bitmembers;
+  
   BOOST_FOREACH(the32bits, d_get32bitpointers) {
-    ret += the32bits.first + "\t" + lexical_cast<string>(*the32bits.second) + "\n";
+    ret.insert(make_pair(the32bits.first, lexical_cast<string>(*the32bits.second)));
   }
   BOOST_FOREACH(the64bits, d_get64bitpointers) {
-    ret += the64bits.first + "\t" + lexical_cast<string>(*the64bits.second) + "\n";
+    ret.insert(make_pair(the64bits.first, lexical_cast<string>(*the64bits.second)));
   }
-  BOOST_FOREACH(the32bitmembers, d_get32bitmembers) {
-    ret += the32bitmembers.first + "\t" + lexical_cast<string>(the32bitmembers.second()) + "\n";
+  BOOST_FOREACH(the32bitmembers, d_get32bitmembers) { 
+    if(the32bitmembers.first == "cache-bytes" || the32bitmembers.first=="packetcache-bytes")
+      continue; // too slow for 'get-all'
+    ret.insert(make_pair(the32bitmembers.first, lexical_cast<string>(the32bitmembers.second())));
+  }
+  return ret;
+}
+
+string getAllStats()
+{
+  typedef map<string, string> varmap_t;
+  varmap_t varmap = getAllStatsMap();
+  string ret;
+  BOOST_FOREACH(varmap_t::value_type& tup, varmap) {
+    ret += tup.first + "\t" + tup.second +"\n";
   }
   return ret;
 }
@@ -137,6 +153,33 @@ static uint64_t* pleaseDump(int fd)
   return new uint64_t(t_RC->doDump(fd) + dumpNegCache(t_sstorage->negcache, fd));
 }
 
+static uint64_t* pleaseDumpNSSpeeds(int fd)
+{
+  return new uint64_t(t_RC->doDumpNSSpeeds(fd));
+}
+
+template<typename T>
+string doDumpNSSpeeds(T begin, T end)
+{
+  T i=begin;
+  string fname;
+
+  if(i!=end)
+    fname=*i;
+
+  int fd=open(fname.c_str(), O_CREAT | O_EXCL | O_WRONLY, 0660);
+  if(fd < 0)
+    return "Error opening dump file for writing: "+string(strerror(errno))+"\n";
+  uint64_t total = 0;
+  try {
+    total = broadcastAccFunction<uint64_t>(boost::bind(pleaseDumpNSSpeeds, fd));
+  }
+  catch(...){}
+
+  close(fd);
+  return "dumped "+lexical_cast<string>(total)+" records\n";
+}
+
 template<typename T>
 string doDumpCache(T begin, T end)
 {
@@ -179,11 +222,12 @@ string doDumpEDNSStatus(T begin, T end)
 
 uint64_t* pleaseWipeCache(const std::string& canon)
 {
-  return new uint64_t(t_RC->doWipeCache(canon));
+  // clear packet cache too
+  return new uint64_t(t_RC->doWipeCache(canon) + t_packetCache->doWipePacketCache(canon));
 }
 
 
-static uint64_t* pleaseWipeAndCountNegCache(const std::string& canon)
+uint64_t* pleaseWipeAndCountNegCache(const std::string& canon)
 {
   uint64_t res = t_sstorage->negcache.count(tie(canon));
   pair<SyncRes::negcache_t::iterator, SyncRes::negcache_t::iterator> range=t_sstorage->negcache.equal_range(tie(canon));
@@ -204,7 +248,36 @@ string doWipeCache(T begin, T end)
   return "wiped "+lexical_cast<string>(count)+" records, "+lexical_cast<string>(countNeg)+" negative records\n";
 }
 
-#ifndef WIN32
+template<typename T>
+string doSetCarbonServer(T begin, T end)
+{
+  Lock l(&g_carbon_config_lock);
+  if(begin==end) {
+    ::arg().set("carbon-server").clear();
+    return "cleared carbon-server setting\n";
+  }
+  string ret;
+  ::arg().set("carbon-server")=*begin;
+  ret="set carbon-server to '"+::arg()["carbon-server"]+"'\n";
+  ++begin;
+  if(begin != end) {
+    ::arg().set("carbon-ourname")=*begin;
+    ret+="set carbon-ourname to '"+*begin+"'\n";
+  }
+  return ret;
+}
+
+
+template<typename T>
+string setMinimumTTL(T begin, T end)
+{
+  if(end-begin != 1) 
+    return "Need to supply new minimum TTL number\n";
+  SyncRes::s_minimumTTL = atoi(begin->c_str());
+  return "New minimum TTL: " + lexical_cast<string>(SyncRes::s_minimumTTL) + "\n";
+}
+
+
 static uint64_t getSysTimeMsec()
 {
   struct rusage ru;
@@ -218,7 +291,6 @@ static uint64_t getUserTimeMsec()
   getrusage(RUSAGE_SELF, &ru);
   return (ru.ru_utime.tv_sec*1000ULL + ru.ru_utime.tv_usec/1000);
 }
-#endif
 
 static uint64_t calculateUptime()
 {
@@ -273,6 +345,16 @@ uint64_t getNegCacheSize()
   return broadcastAccFunction<uint64_t>(pleaseGetNegCacheSize);
 }
 
+uint64_t* pleaseGetFailedHostsSize()
+{
+  uint64_t tmp=t_sstorage->fails.size();
+  return new uint64_t(tmp);
+}
+uint64_t getFailedHostsSize()
+{
+  return broadcastAccFunction<uint64_t>(pleaseGetFailedHostsSize);
+}
+
 uint64_t* pleaseGetNsSpeedsSize()
 {
   return new uint64_t(t_sstorage->nsSpeeds.size());
@@ -298,11 +380,28 @@ uint64_t* pleaseGetCacheSize()
   return new uint64_t(t_RC->size());
 }
 
+uint64_t* pleaseGetCacheBytes()
+{
+  return new uint64_t(t_RC->bytes());
+}
+
+
 uint64_t doGetCacheSize()
 {
   return broadcastAccFunction<uint64_t>(pleaseGetCacheSize);
 }
 
+uint64_t doGetAvgLatencyUsec()
+{
+  return (uint64_t) g_stats.avgLatencyUsec;
+}
+
+
+uint64_t doGetCacheBytes()
+{
+  return broadcastAccFunction<uint64_t>(pleaseGetCacheBytes);
+}
+
 uint64_t* pleaseGetCacheHits()
 {
   return new uint64_t(t_RC->cacheHits);
@@ -324,18 +423,28 @@ uint64_t doGetCacheMisses()
 }
 
 
-
-
 uint64_t* pleaseGetPacketCacheSize()
 {
   return new uint64_t(t_packetCache->size());
 }
 
+uint64_t* pleaseGetPacketCacheBytes()
+{
+  return new uint64_t(t_packetCache->bytes());
+}
+
+
 uint64_t doGetPacketCacheSize()
 {
   return broadcastAccFunction<uint64_t>(pleaseGetPacketCacheSize);
 }
 
+uint64_t doGetPacketCacheBytes()
+{
+  return broadcastAccFunction<uint64_t>(pleaseGetPacketCacheBytes);
+}
+
+
 uint64_t* pleaseGetPacketCacheHits()
 {
   return new uint64_t(t_packetCache->d_hits);
@@ -356,21 +465,38 @@ uint64_t doGetPacketCacheMisses()
   return broadcastAccFunction<uint64_t>(pleaseGetPacketCacheMisses);
 }
 
+uint64_t doGetMallocated()
+{
+  // this turned out to be broken
+/*  struct mallinfo mi = mallinfo();
+  return mi.uordblks; */
+  return 0;
+}
 
+extern ResponseStats g_rs;
+
+bool RecursorControlParser::s_init;
 RecursorControlParser::RecursorControlParser()
 {
+  if(s_init)
+    return;
+  s_init=true;
+
   addGetStat("questions", &g_stats.qcounter);
+  addGetStat("ipv6-questions", &g_stats.ipv6qcounter);
   addGetStat("tcp-questions", &g_stats.tcpqcounter);
 
   addGetStat("cache-hits", doGetCacheHits);
   addGetStat("cache-misses", doGetCacheMisses); 
   addGetStat("cache-entries", doGetCacheSize); 
+  addGetStat("cache-bytes", doGetCacheBytes); 
   
   addGetStat("packetcache-hits", doGetPacketCacheHits);
   addGetStat("packetcache-misses", doGetPacketCacheMisses); 
   addGetStat("packetcache-entries", doGetPacketCacheSize); 
+  addGetStat("packetcache-bytes", doGetPacketCacheBytes); 
   
-  
+  addGetStat("malloc-bytes", doGetMallocated);
   
   addGetStat("servfail-answers", &g_stats.servFails);
   addGetStat("nxdomain-answers", &g_stats.nxDomains);
@@ -389,7 +515,7 @@ RecursorControlParser::RecursorControlParser()
   addGetStat("answers100-1000", &g_stats.answers100_1000);
   addGetStat("answers-slow", &g_stats.answersSlow);
 
-  addGetStat("qa-latency", &g_stats.avgLatencyUsec);
+  addGetStat("qa-latency", doGetAvgLatencyUsec);
   addGetStat("unexpected-packets", &g_stats.unexpectedCount);
   addGetStat("case-mismatches", &g_stats.caseMismatchCount);
   addGetStat("spoof-prevents", &g_stats.spoofCount);
@@ -398,13 +524,16 @@ RecursorControlParser::RecursorControlParser()
 
   addGetStat("resource-limits", &g_stats.resourceLimits);
   addGetStat("over-capacity-drops", &g_stats.overCapacityDrops);
+  addGetStat("policy-drops", &g_stats.policyDrops);
   addGetStat("no-packet-error", &g_stats.noPacketError);
   addGetStat("dlg-only-drops", &SyncRes::s_nodelegated);
+  addGetStat("max-mthread-stack", &g_stats.maxMThreadStackUsage);
   
   addGetStat("negcache-entries", boost::bind(getNegCacheSize));
   addGetStat("throttle-entries", boost::bind(getThrottleSize)); 
 
   addGetStat("nsspeeds-entries", boost::bind(getNsSpeedsSize));
+  addGetStat("failed-host-entries", boost::bind(getFailedHostsSize));
 
   addGetStat("concurrent-queries", boost::bind(getConcurrentQueries)); 
   addGetStat("outgoing-timeouts", &SyncRes::s_outgoingtimeouts);
@@ -426,11 +555,9 @@ RecursorControlParser::RecursorControlParser()
 
   addGetStat("uptime", calculateUptime);
 
-#ifndef WIN32
   //  addGetStat("query-rate", getQueryRate);
   addGetStat("user-msec", getUserTimeMsec);
   addGetStat("sys-msec", getSysTimeMsec);
-#endif
 }
 
 static void doExitGeneric(bool nicely)
@@ -455,6 +582,8 @@ static void doExit()
 
 static void doExitNicely()
 {
+  //extern void printCallers();
+  // printCallers();
   doExitGeneric(true);
 }
 
@@ -477,7 +606,7 @@ string doTopRemotes()
       counts[*i]++;
     }
 
-  typedef multimap<int, ComboAddress> rcounts_t;
+  typedef std::multimap<int, ComboAddress> rcounts_t;
   rcounts_t rcounts;
   
   for(counts_t::const_iterator i=counts.begin(); i != counts.end(); ++i)
@@ -511,6 +640,32 @@ string RecursorControlParser::getAnswer(const string& question, RecursorControlP
   string cmd=toLower(words[0]);
   vector<string>::const_iterator begin=words.begin()+1, end=words.end();
 
+  // should probably have a smart dispatcher here, like auth has
+  if(cmd=="help")
+    return
+"current-queries                  show currently active queries\n"
+"dump-cache <filename>            dump cache contents to the named file\n"
+"dump-edns[status] <filename>     dump EDNS status to the named file\n"
+"dump-nsspeeds <filename>         dump nsspeeds statistics to the named file\n"
+"get [key1] [key2] ..             get specific statistics\n"
+"get-all                          get all statistics\n"
+"get-parameter [key1] [key2] ..   get configuration parameters\n"
+"get-qtypelist                    get QType statistics\n"
+"                                 notice: queries from cache aren't being counted yet\n"
+"help                             get this list\n"
+"ping                             check that all threads are alive\n"
+"quit                             stop the recursor daemon\n"
+"quit-nicely                      stop the recursor daemon nicely\n"
+"reload-acls                      reload ACLS\n"
+"reload-lua-script [filename]     (re)load Lua script\n"
+"reload-zones                     reload all auth and forward zones\n"
+"set-minimum-ttl value            set mininum-ttl-override\n"
+"set-carbon-server                set a carbon server for telemetry\n"
+"trace-regex [regex]              emit resolution trace for matching queries (empty regex to clear trace)\n"
+"top-remotes                      show top remotes\n"
+"unload-lua-script                unload Lua script\n"
+"wipe-cache domain0 [domain1] ..  wipe domain data from cache\n";
+
   if(cmd=="get-all")
     return getAllStats();
 
@@ -520,7 +675,6 @@ string RecursorControlParser::getAnswer(const string& question, RecursorControlP
   if(cmd=="get-parameter") 
     return doGetParameter(begin, end);
 
-
   if(cmd=="quit") {
     *command=&doExit;
     return "bye\n";
@@ -529,8 +683,7 @@ string RecursorControlParser::getAnswer(const string& question, RecursorControlP
   if(cmd=="quit-nicely") {
     *command=&doExitNicely;
     return "bye nicely\n";
-  }
-  
+  }  
 
   if(cmd=="dump-cache") 
     return doDumpCache(begin, end);
@@ -538,13 +691,21 @@ string RecursorControlParser::getAnswer(const string& question, RecursorControlP
   if(cmd=="dump-ednsstatus" || cmd=="dump-edns") 
     return doDumpEDNSStatus(begin, end);
 
+  if(cmd=="dump-nsspeeds")
+    return doDumpNSSpeeds(begin, end);
 
-  if(cmd=="wipe-cache") 
+  if(cmd=="wipe-cache" || cmd=="flushname") 
     return doWipeCache(begin, end);
 
   if(cmd=="reload-lua-script") 
     return doQueueReloadLuaScript(begin, end);
 
+  if(cmd=="set-carbon-server") 
+    return doSetCarbonServer(begin, end);
+
+  if(cmd=="trace-regex") 
+    return doTraceRegex(begin, end);
+
   if(cmd=="unload-lua-script") {
     vector<string> empty;
     empty.push_back(string());
@@ -555,10 +716,16 @@ string RecursorControlParser::getAnswer(const string& question, RecursorControlP
     try {
       parseACLs();
     } 
-    catch(exception& e) 
+    catch(std::exception& e) 
     {
+      L<<Logger::Error<<"reloading ACLs failed (Exception: "<<e.what()<<")"<<endl;
       return e.what() + string("\n");
     }
+    catch(PDNSException& ae)
+    {
+      L<<Logger::Error<<"reloading ACLs failed (PDNSException: "<<ae.reason<<")"<<endl;
+      return ae.reason + string("\n");
+    }
     return "ok\n";
   }
 
@@ -577,5 +744,13 @@ string RecursorControlParser::getAnswer(const string& question, RecursorControlP
     return reloadAuthAndForwards();
   }
 
-  return "Unknown command '"+cmd+"'\n";
+  if(cmd=="set-minimum-ttl") {
+    return setMinimumTTL(begin, end);
+  }
+  
+  if(cmd=="get-qtypelist") {
+    return g_rs.getQTypeReport();
+  }
+  
+  return "Unknown command '"+cmd+"', try 'help'\n";
 }
