@@ -6,6 +6,10 @@
     it under the terms of the GNU General Public License version 2 as 
     published by the Free Software Foundation
 
+    Additionally, the license of this program contains a special
+    exception which allows to distribute the program in binary form when
+    it is linked against OpenSSL.
+
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@@ -21,12 +25,13 @@
 #include <fcntl.h>
 #include <unistd.h>
 #include <stdlib.h> 
-#include "ahuexception.hh"
+#include "pdnsexception.hh"
 #include "logger.hh"
 #include "misc.hh"
 #include <pwd.h>
 #include <grp.h>
 #include <sys/types.h>
+#include <sys/select.h>
 
 #ifdef NEED_INET_NTOP_PROTO
 extern "C" {
@@ -44,7 +49,37 @@ int Utility::closesocket( Utility::sock_t socket )
   if(ret < 0 && errno == ECONNRESET) // see ticket 192, odd BSD behaviour
     return 0;
   if(ret < 0) 
-    throw AhuException("Error closing socket: "+stringerror());
+    throw PDNSException("Error closing socket: "+stringerror());
+  return ret;
+}
+
+// Connects to socket with timeout
+int Utility::timed_connect( Utility::sock_t sock,
+    const sockaddr *addr,
+    Utility::socklen_t sockaddr_size,
+    int timeout_sec,
+    int timeout_usec )
+{
+  fd_set set;
+  struct timeval timeout;
+  int ret;
+
+  timeout.tv_sec = timeout_sec;
+  timeout.tv_usec = timeout_usec;
+
+  FD_ZERO(&set);
+  FD_SET(sock, &set);
+
+  setNonBlocking(sock);
+
+  if ((ret = connect (sock, addr, sockaddr_size)) < 0) {
+    if (errno != EINPROGRESS)
+      return ret;
+  }
+
+  ret = select(sock + 1, NULL, &set, NULL, &timeout);
+  setBlocking(sock);
+
   return ret;
 }
 
@@ -88,8 +123,8 @@ void Utility::usleep(unsigned long usec)
 }
 
 
-// Drops the program's privileges.
-void Utility::dropPrivs( int uid, int gid )
+// Drops the program's group privileges.
+void Utility::dropGroupPrivs( int uid, int gid )
 {
   if(gid) {
     if(setgid(gid)<0) {
@@ -99,13 +134,26 @@ void Utility::dropPrivs( int uid, int gid )
     else
       theL()<<Logger::Info<<"Set effective group id to "<<gid<<endl;
 
-    if(setgroups(0, NULL)<0) {
-      theL()<<Logger::Critical<<"Unable to drop supplementary gids: "<<stringerror()<<endl;
-      exit(1);
+    struct passwd *pw=getpwuid(uid);
+    if(!pw) {
+      theL()<<Logger::Warning<<"Unable to determine user name for uid "<<uid<<endl;
+      if (setgroups(0, NULL)<0) {
+        theL()<<Logger::Critical<<"Unable to drop supplementary gids: "<<stringerror()<<endl;
+        exit(1);
+      }
+    } else {
+      if (initgroups(pw->pw_name, gid)<0) {
+        theL()<<Logger::Critical<<"Unable to set supplementary groups: "<<stringerror()<<endl;
+        exit(1);
+      }
     }
-
   }
+}
 
+
+// Drops the program's user privileges.
+void Utility::dropUserPrivs( int uid )
+{
   if(uid) {
     if(setuid(uid)<0) {
       theL()<<Logger::Critical<<"Unable to set effective user id to "<<uid<<":  "<<stringerror()<<endl;
@@ -229,7 +277,7 @@ time_t Utility::timegm(struct tm *const t)
   if (t->tm_sec>60) { t->tm_min += t->tm_sec/60; t->tm_sec%=60; }
   if (t->tm_min>60) { t->tm_hour += t->tm_min/60; t->tm_min%=60; }
   if (t->tm_hour>60) { t->tm_mday += t->tm_hour/60; t->tm_hour%=60; }
-  if (t->tm_mon>12) { t->tm_year += t->tm_mon/12; t->tm_mon%=12; }
+  if (t->tm_mon>11) { t->tm_year += t->tm_mon/12; t->tm_mon%=12; }
  
   while (t->tm_mday>spm[1+t->tm_mon]) {
     if (t->tm_mon==1 && isleap(t->tm_year+1900)) {
@@ -246,9 +294,9 @@ time_t Utility::timegm(struct tm *const t)
   /* Days since 1970 is 365 * number of years + number of leap years since 1970 */
   day  = years * 365 + (years + 1) / 4;
 
-  /* After 2100 we have to substract 3 leap years for every 400 years
+  /* After 2100 we have to subtract 3 leap years for every 400 years
      This is not intuitive. Most mktime implementations do not support
-     dates after 2059, anyway, so we might leave this out for it's
+     dates after 2059, anyway, so we might leave this out for its
      bloat. */
   if ((years -= 131) >= 0) {
     years /= 100;
@@ -269,3 +317,30 @@ time_t Utility::timegm(struct tm *const t)
   return ((day + t->tm_hour) * i + t->tm_min) * i + t->tm_sec;
 }
 
+// we have our own gmtime_r because the one in GNU libc violates POSIX/SuS
+// by supporting leap seconds when TZ=right/UTC
+void Utility::gmtime_r(const time_t *timer, struct tm *tmbuf) {
+
+  int monthdays[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
+  int days = *timer / 86400;
+  int leapdays = (days + 671) / 1461;
+  int leapydays = (days + 365) / 1461;
+
+  tmbuf->tm_hour = *timer / 3600 % 24;
+  tmbuf->tm_min = *timer / 60 % 60;
+  tmbuf->tm_sec = *timer % 60;
+
+  tmbuf->tm_year = (days - leapdays) / 365 + 70;
+  tmbuf->tm_yday = days - leapydays - (tmbuf->tm_year - 70) * 365 + 1;
+
+  tmbuf->tm_mon = 0;
+  tmbuf->tm_mday = tmbuf->tm_yday;
+  monthdays[1] += isleap(tmbuf->tm_year + 1900);
+  while (monthdays[tmbuf->tm_mon] < tmbuf->tm_mday) {
+    tmbuf->tm_mday -= monthdays[tmbuf->tm_mon];
+    tmbuf->tm_mon++;
+  }
+
+  tmbuf->tm_wday = (days + 4) % 7; // Day 0 is magic thursday ;)
+  tmbuf->tm_isdst = 0;
+}
